/*
https://github.com/AmberAbernathy/Color_Quantization/blob/main/test_km_algs.cpp

Online K-Means (MacQueen, 1967)
Incremental Online K-Means (Abernathy & Celebi, 2022)

Authors: Amber Abernathy & M. Emre Celebi

Contact email: ecelebi@uca.edu

If you find this program useful, please cite:
A. D. Abernathy and M. E. Celebi, 
The Incremental Online K-Means Clustering Algorithm 
and Its Application to Color Quantization,
Expert Systems with Applications,
in press, https://doi.org/10.1016/j.eswa.2022.117927, 2022.
*/

#include <climits>
#include <math.h>
#include "header.hpp"

typedef unsigned char uchar;
typedef unsigned int uint;
typedef unsigned long ulong;

struct RGB_Cluster
{
 int size;
 Color center;
};

/* Max. # colors that can be requested */
#define MAX_NUM_COLORS 256

/* 
  Powers of two for 0, 1, ..., 16. Note that 2^16 must equal MAX_NUM_COLORS.
  If you want to quantize to more than MAX_NUM_COLORS colors, extend the POW2 
  array accordingly.
 */
static inline int POW2[] = { 1, 2, 4, 8, 16, 32, 64, 128, 256 };

/*
  Function to generate two quasirandom numbers from 
  a 2D Sobol sequence. Adapted from Numerical Recipies 
  in C. Upon return, X and Y fall in [0,1).
 */

#define MAX_BIT 30

void 
sob_seq ( double *x, double *y )
{
 int j, k, l;
 ulong i, im, ipp;
 static double fac;
 static int init = 0;
 static ulong ix1, ix2;
 static ulong in, *iu[2 * MAX_BIT + 1];
 static ulong mdeg[3] = { 0, 1, 2 };
 static ulong ip[3] = { 0, 0, 1 };
 static ulong iv[2 * MAX_BIT + 1] =
 { 0, 1, 1, 1, 1, 1, 1, 3, 1, 3, 3, 1, 1, 5, 7, 7, 3, 3, 5, 15, 11, 5, 15, 13, 9 };

 if ( !init )
  {
   init = 1;
   for ( j = 1, k = 0; j <= MAX_BIT; j++, k += 2 )
    {
     iu[j] = &iv[k];
    }

   for ( k = 1; k <= 2; k++ )
    {
     for ( j = 1; j <= ( int ) mdeg[k]; j++ )
      {
       iu[j][k] <<= ( MAX_BIT - j );
      }

     for ( j = mdeg[k] + 1; j <= MAX_BIT; j++ )
      {
       ipp = ip[k];
       i = iu[j - mdeg[k]][k];
       i ^= ( i >> mdeg[k] );

       for ( l = mdeg[k] - 1; l >= 1; l-- )
        {
 	 if ( ipp & 1 )
 	  {
  	   i ^= iu[j - l][k];
	  }

	 ipp >>= 1;
	}

       iu[j][k] = i;
      }
    }

   fac = 1.0 / ( 1L << MAX_BIT );
   in = 0;
  }

 im = in;
 for ( j = 1; j <= MAX_BIT; j++ )
  {
   if ( ! ( im & 1 ) )
    {
     break;
    }

   im >>= 1;
  }

 im = (j - 1) * 2;
 *x = (ix1 ^= iv[im + 1]) * fac;
 *y = (ix2 ^= iv[im + 2]) * fac;
 in++;
}

#undef MAX_BIT

/*
  Function to determine if an integer is a power of 2:
  http://graphics.stanford.edu/~seander/bithacks.html#DetermineIfPowerOf2
*/

bool
is_pow2 ( const int x )
{ 
 uint ux = ( uint ) x;

 return ux && !( ux & ( ux - 1 ) );
}

/* 
  Online K-Means Algorithm:
  S. Thompson, M. E. Celebi, and K. H. Buck, 
  Fast Color Quantization Using MacQueen’s K-Means Algorithm, 
  Journal of Real-Time Image Processing, 
  17(5): 1609-1624, 2020.  

  Notes:
        1) LR_EXP: Learning rate exponent (must be in [0.5, 1])
        2) SAMPLE_RATE: Fraction of the input pixels (must be in (0, 1]) 
	   used during the clustering process.
	3) CLUST: When the function is called, CLUST represents the initial 
	   centers. Upon return, CLUST represents the final centers.
 */

void 
online_kmeans ( const Image *img, const int num_colors, const double lr_exp, 
		const double sample_rate, RGB_Cluster *clust )
{
 int min_dist_index;
 int old_size, new_size;
 int row_idx, col_idx;
 int num_samples;
 double sob_x, sob_y;
 double del_red, del_green, del_blue;
 double dist, min_dist;
 double learn_rate;
 Color rand_pixel;

 if ( lr_exp < 0.5 || lr_exp > 1. )
  {
   fprintf ( stderr, "Learning rate exponent (%g) must be in [0.5, 1]\n", lr_exp );
   exit ( EXIT_FAILURE );
  }
 else if ( sample_rate <= 0.0 || sample_rate > 1. )
  {
   fprintf ( stderr, "Sampling rate (%g) must be in (0, 1]\n", sample_rate );
   exit ( EXIT_FAILURE );
  }

 num_samples = ( int ) ( sample_rate * (img->width*img->height) + 0.5 ); /* round */
	
 for ( int i = 0; i < num_samples; i++ ) 
  {
   /* Sample the image quasirandomly based on a Sobol' sequence */
   sob_seq ( &sob_x, &sob_y );

   /* Find the corresponding row/column indices */
   row_idx = ( int ) ( sob_y * img->height + 0.5 ); /* round */
   if ( row_idx == img->height )
    {
     row_idx--;
    }

   col_idx = ( int ) ( sob_x * img->width + 0.5 ); /* round */
   if ( col_idx == img->width )
    {
     col_idx--;
    }

   rand_pixel = img->data[row_idx * img->width + col_idx];

   /* Find the nearest center */
   min_dist = MAX_RGB_DIST;
   min_dist_index = -INT_MAX;
   for ( int j = 0; j < num_colors; j++ ) 
    {
     del_red = clust[j].center.r - rand_pixel.r;
     del_green = clust[j].center.g - rand_pixel.g;
     del_blue = clust[j].center.b - rand_pixel.b;

     dist = del_red * del_red + del_green * del_green + del_blue * del_blue;
     if ( dist < min_dist ) 
      {
       min_dist = dist;
       min_dist_index = j;
      }
    }

   /* Update the size of the nearest cluster */
   old_size = clust[min_dist_index].size;
   new_size = old_size + 1;

   /* Compute the learning rate */
   learn_rate = pow ( new_size, -lr_exp );	

   /* Update the center of the nearest cluster */
   clust[min_dist_index].center.r += learn_rate * 
	                               ( rand_pixel.r - clust[min_dist_index].center.r );
   clust[min_dist_index].center.g += learn_rate * 
	                                 ( rand_pixel.g - clust[min_dist_index].center.g );
   clust[min_dist_index].center.b += learn_rate * 
	                                ( rand_pixel.b - clust[min_dist_index].center.b );
   clust[min_dist_index].size = new_size;
  }
}

/* Function to compute the centroid of an image */

RGB_Cluster 
compute_centroid ( const Image *img ) 
{
 double sum_red = 0.0, sum_green = 0.0, sum_blue = 0.0;
 Color pixel;
 RGB_Cluster centroid;

 for (int i = 0; i < (img->width*img->height); i++) 
  {
   pixel = img->data[i];
   sum_red += pixel.r;
   sum_green += pixel.g;
   sum_blue += pixel.b;
  }

 centroid.center.r = sum_red / (img->width*img->height);
 centroid.center.g = sum_green / (img->width*img->height);
 centroid.center.b = sum_blue / (img->width*img->height);

 return centroid;
}

/* 
  Incremental Online K-Means Algorithm:

  A. D. Abernathy and M. E. Celebi, 
  The Incremental Online K-Means Clustering Algorithm 
  and Its Application to Color Quantization,
  Expert Systems with Applications,
  accepted for publication, 2022.
  
  Notes:
        1) NUM_COLORS must be a power of 2 (otherwise the code must be 
	   modified slightly, see Abernathy & Celebi, 2022).
        2) LR_EXP: Learning rate exponent (must be in [0.5, 1])
        3) SAMPLE_RATE: Fraction of the input pixels (must be in (0, 1]) 
	   used during the clustering process.
 */

std::vector<Color>
inc_online_kmeans ( const Image *img, const int num_colors, 
		    const double lr_exp, const double sample_rate )
{
 int index, num_splits;
 Color pixel;
 RGB_Cluster *tmp_clust, *clust;
 
 if ( !is_pow2 ( num_colors ) )
  {
   fprintf ( stderr, "Number of colors (%d) must be a power of 2!\n", num_colors );
   exit ( EXIT_FAILURE );
  }

 /* Compute log2 ( num_colors ) */
 num_splits = ( int ) ( log ( num_colors ) / log ( 2 ) + 0.5 ); /* round */
 
 tmp_clust = ( RGB_Cluster * ) malloc ( ( 2 * num_colors - 1 ) * sizeof ( RGB_Cluster ) );
 clust = ( RGB_Cluster * ) malloc ( num_colors * sizeof ( RGB_Cluster ) );

 /* Set first center to be the dataset centroid */
 tmp_clust[0] = compute_centroid ( img );
 tmp_clust[0].size = 0;

 for ( int t = 0; t < num_splits; t++ )
  {
   for ( int n = POW2[t] - 1; n < POW2[t + 1] - 1; n++ )
    {
     /* Split c_n into c_{2n + 1} and c_{2n + 2} */
     pixel = tmp_clust[n].center;

     /* Left child */
     index = 2 * n + 1;
     tmp_clust[index].center.r = pixel.r;
     tmp_clust[index].center.g = pixel.g;
     tmp_clust[index].center.b = pixel.b;
     tmp_clust[index].size = 0;
	
     /* Right child */
     index++;
     tmp_clust[index].center.r = pixel.r;
     tmp_clust[index].center.g = pixel.g;
     tmp_clust[index].center.b = pixel.b;
     tmp_clust[index].size = 0;
    }

   /* Refine the new centers using online k-means */
   online_kmeans ( img, POW2[t + 1], lr_exp, sample_rate, 
		   tmp_clust + POW2[t + 1] - 1 );
  }
	
 /* Last NUM_COLORS centers are the final centers */
 for ( int j = 0; j < num_colors; j++ ) 
  {
   clust[j].center.r = tmp_clust[j + num_colors - 1].center.r;
   clust[j].center.g = tmp_clust[j + num_colors - 1].center.g;
   clust[j].center.b = tmp_clust[j + num_colors - 1].center.b;
  }

 free (tmp_clust);

 std::vector<Color> result(num_colors);
 for (int i=0; i<num_colors; i++)
  result[i] = clust[i].center;

 free(clust);
 return result;
}