/*
 * Copyright (C) 2024 sillysagiri
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include <cassert>
#include <cstdint>
#include <cstdlib>
#include <filesystem>
#include <fstream>
#include <stdexcept>
#include <format>
#include <iostream>
#include <vector>

#define STB_IMAGE_IMPLEMENTATION
#define STB_IMAGE_WRITE_IMPLEMENTATION

#include <argparse/argparse.hpp>
#include "header.hpp"
#include "Timer.h"
#include "fastlz.h"
#include "murmurhash.h"

namespace fs = std::filesystem;

int main(int argc, char **argv)
{
	srand(time(0));
	argparse::ArgumentParser program("sillyimage", "1.3");
	Options opt;

	// Register argument
	{
		program.add_argument("-i", "--input")
			.help("input files")
			.append()
			.required()
			.store_into(opt.path_input);

		program.add_argument("-o", "--out")
			.help("output directory")
			.required()
			.store_into(opt.path_output);

		program.add_argument("-f", "--format")
			.help("texture format { rgb256, rgb16, indexed4, indexed16, indexed256, indexed32a8, indexed8a32 }")
			.default_value<std::string>("rgb16")
			.choices("rgb256", "rgb16", "indexed4", "indexed16", "indexed256", "indexed32a8", "indexed8a32")
			.nargs(1)
			.store_into(opt.format);
	}

	// Parse argument
	{
		try
		{
			program.parse_args(argc, argv);

			// input list
			for (auto &i : opt.path_input)
				if (i.empty() || !fs::exists(i)) 
					throw std::runtime_error(std::format("file doesnt exsist: \"{}\"", i));

			// output directory
			if (opt.path_output.empty() || !fs::is_directory(opt.path_output))
				throw std::runtime_error(std::format("invalid output directory: \"{}\"", opt.path_output));

		}
		catch(const std::exception &e)
		{
			std::cerr << std::format("Error: {}\n\n", e.what()) << program << "\n";
			return 1;
		}
	}

	// Convert
	try {

		if (opt.format == "rgb256")
		{
			for(auto &input : opt.path_input)
			{
				Image img;
				img.Load(input);

				Metadata meta;
				meta.format = Format::Format_RGB_256;
				meta.width = img.width;
				meta.height = img.height;
				meta.palette_count = 0;
				meta.palette_hash = 0;
				meta.original_size = img.width*img.height*4;

				uint8_t compress[uint64_t(meta.original_size*1.5)];
				meta.compress_size = fastlz_compress_level(1, img.data.data(), meta.original_size, compress);

				auto output = fs::path(opt.path_output) / std::format("{}.sillyimg", fs::path(input).stem().string());
				std::ofstream out(output, std::ios::binary);
				out.write(reinterpret_cast<const char*>(&meta), sizeof(meta));
				out.write(reinterpret_cast<const char*>(compress), meta.compress_size);
				out.close();
			}
		}

		if (opt.format == "rgb16")
		{
			for(auto &input : opt.path_input)
			{
				Image img;
				img.Load(input);
				uint16_t img16[img.width*img.height];

				// convert into RGB16 color
				for (int i=0; i<img.width*img.height; i++)
					img16[i] = img.data[i].toRGB16();

				Metadata meta;
				meta.format = Format::Format_RGB_16;
				meta.width = img.width;
				meta.height = img.height;
				meta.palette_count = 0;
				meta.palette_hash = 0;
				meta.original_size = img.width*img.height*2;

				uint8_t compress[uint64_t(meta.original_size*1.5)];
				meta.compress_size = fastlz_compress_level(1, img16, meta.original_size, compress);

				auto output = fs::path(opt.path_output) / std::format("{}.sillyimg", fs::path(input).stem().string());
				std::ofstream out(output, std::ios::binary);
				out.write(reinterpret_cast<const char*>(&meta), sizeof(meta));
				out.write(reinterpret_cast<const char*>(compress), meta.compress_size);
				out.close();
			}
		}

		if (opt.format == "indexed4" ||
				opt.format == "indexed16" ||
				opt.format == "indexed256" ||
				opt.format == "indexed32a8" ||
				opt.format == "indexed8a32")
		{
			uint16_t numcol;
			uint8_t format;

			if (opt.format == "indexed4")
			{
				numcol = 4;
				format = Format_INDEXED_4;
			}
			else if (opt.format == "indexed16")
			{
				numcol = 16;
				format = Format_INDEXED_16;
			}
			else if (opt.format == "indexed256")
			{
				numcol = 256;
				format = Format_INDEXED_256;
			}
			else if (opt.format == "indexed32a8")
			{
				numcol = 32;
				format = Format_INDEXED_32A8;
			}
			else if (opt.format == "indexed8a32")
			{
				numcol = 8;
				format = Format_INDEXED_8A32;
			}

			std::vector<Color> palette;
			std::vector<Image> images(opt.path_input.size());
			
			for (int i=0; i<images.size(); i++)
				images[i].Load(opt.path_input[i]);

			if (images.size() > 1)
			{
				int combined_size = 0;
				Image combined;
				
				// pre-calclate size
				for (auto &image : images)
					combined_size += image.data.size();

				combined.data.reserve(combined_size);

				for (const auto& image : images)
					combined.data.insert(combined.data.end(), image.data.begin(), image.data.end());

				palette = QuantizePalette(numcol, combined);
			}
			else palette = QuantizePalette(numcol, images[0]);

			uint16_t palette16[palette.size()];
			for (int i=0; i<palette.size(); i++)
				palette16[i] = palette[i].toRGB16();

			uint32_t hash = murmurhash(reinterpret_cast<const char*>(palette16), sizeof(uint16_t) * palette.size(), 0);

			for (int i=0; i<images.size(); i++)
			{
				std::vector<uint8_t> remap = RemapImage(images[i], palette, format);
				// std::vector<uint8_t> remap = {1,2,3,4,54,5,6,7,78,23,34,32,32,12,5,34};

				Metadata meta;
				meta.format = format;
				meta.width = images[i].width;
				meta.height = images[i].height;
				meta.palette_count = palette.size();
				meta.palette_hash = hash;
				meta.original_size = remap.size();

				uint8_t compress[uint64_t(meta.original_size*1.5)];
				meta.compress_size = fastlz_compress_level(1, remap.data(), meta.original_size, compress);

				auto output = fs::path(opt.path_output) / std::format("{}.sillyimg", fs::path(opt.path_input[i]).stem().string());
				// std::ofstream out(output, std::ios::binary);
				
				// out.write(reinterpret_cast<const char*>(&meta), sizeof(meta));
				// out.write(reinterpret_cast<const char*>(palette16), sizeof(uint16_t) * palette.size());
				// out.write(reinterpret_cast<const char*>(compress), meta.compress_size);
				// out.close();
			}
		}
	}
	catch(const std::exception &e)
	{
		std::cerr << std::format("Exception caught: {}\n", e.what());
		return 1;
	}
	catch (...)
	{
		std::cerr << std::format("An unknown exception occurred\n");
		return 1;
	}

	std::cout  << "done\n";

	return 0;
}

// // ----

std::vector<Color> QuantizePalette(const int numColor, const Image &image)
{
	return inc_online_kmeans(&image, numColor);
}

std::vector<uint8_t> RemapImage(const Image &image, const std::vector<Color> &palette, const int &format)
{
	std::vector<uint8_t> remap(image.width*image.height);

	int min_dist_index;
	double del_red, del_green, del_blue;
	double dist, min_dist;

	// TODO: explore more distance algorithm
	for (int i=0; i<image.data.size(); i++)
	{
		const auto &pixel = image.data[i];
		min_dist = MAX_RGB_DIST;
		min_dist_index = -INT_MAX;

		for (int j=0; j<palette.size(); j++) 
		{
			del_red = palette[j].r - pixel.r;
			del_green = palette[j].g - pixel.g;
			del_blue = palette[j].b - pixel.b;

			dist = del_red * del_red + del_green * del_green + del_blue * del_blue;
			if ( dist < min_dist ) 
			{
				min_dist = dist;
				min_dist_index = j;
			}
		}

		remap[i] = min_dist_index;
	}

	// ---

	if (format == Format_INDEXED_4)
	{
		std::vector<uint8_t> packed;
		packed.reserve((remap.size() + 3) / 4); // Round up
		
		for (size_t i = 0; i < remap.size(); ) {
			uint8_t byte = 0;
			byte |= (remap[i] & 0b00000011);
			
			if (++i < remap.size()) byte |= (remap[i] & 0b00000011) << 2;
			if (++i < remap.size()) byte |= (remap[i] & 0b00000011) << 4;
			if (++i < remap.size()) byte |= (remap[i] & 0b00000011) << 6;
			
			packed.push_back(byte);
			i++; // Move to next group
		}
		
		return packed;
	}

	else if (format == Format_INDEXED_16)
	{
		std::vector<uint8_t> packed;
		packed.reserve((remap.size() + 1) / 2); // Round up

		for (int i = 0; i < remap.size(); ) {
			uint8_t byte = 0;
			byte |= (remap[i] & 0b00001111);

			if (++i < remap.size()) {
					byte |= (remap[i] & 0b00001111) << 4;
			}

			packed.push_back(byte);
			i++; // Move to next group
		}

		return packed;
	}

	else if (format == Format_INDEXED_8A32)
	{
		std::vector<uint8_t> packed;
		packed.reserve(remap.size());

		for (size_t i = 0; i < remap.size(); ++i) {
			uint8_t color_index = remap[i] & 0b00000111;  // 3-bit mask
			uint8_t alpha = ((uint8_t)palette[remap[i]].a >> 3) & 0b00011111; // 5-bit mask

			packed.push_back((alpha << 3) | color_index);
		}

		return packed;
	}

	else if (format == Format_INDEXED_32A8)
	{
		std::vector<uint8_t> packed;
		packed.reserve(remap.size());

		for (size_t i = 0; i < remap.size(); ++i) {
			uint8_t color_index = remap[i] & 0b00011111;  // 5-bit mask
			uint8_t alpha = ((uint8_t)palette[remap[i]].a >> 5) & 0b00000111; // 3-bit mask

			packed.push_back((alpha << 5) | color_index);
		}

		return packed;
	}

	return remap;
}

// // ----

void Image::Load(const std::filesystem::path &path)
{
	int comp;
	stbi_uc *img = stbi_load(path.c_str(), &width, &height, &comp, 4);
	if (img == nullptr) throw std::runtime_error(stbi_failure_reason());

	data.resize(width*height);
	for (int i=0; i<width*height; i++) {
		const int offset = i * 4;
		data[i].r = img[offset];
		data[i].g = img[offset+1];
		data[i].b = img[offset+2];
		data[i].a = img[offset+3];
	}

	stbi_image_free(img);
}